#include "renderer3D.hpp"

#include "exceptions.hpp"
#include "vk_enum_string.h"

#include <cstring>

namespace gz::vk {

Renderer3D::Renderer3D(VulkanInstance& instance) : 
    Renderer(instance),
    rLog("renderer3D.log", true, false, "3D-Renderer", Color::BCYAN, true, 100) {
    vk.createCommandBuffers(commandBuffers);
    const size_t vertexCount = 500;
    const size_t indexCount = 1000;
    vk.createVertexBuffer<Vertex2D>(vertexCount, vertexBuffer, vertexBufferMemory, vertexBufferSize);
    vk.createIndexBuffer<uint32_t>(indexCount, indexBuffer, indexBufferMemory, indexBufferSize);
    rLog("Created Renderer3D");
}

void Renderer3D::cleanup() {
    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        vkDestroyBuffer(vk.device, uniformBuffers[i], nullptr);
        vkFreeMemory(vk.device, uniformBuffersMemory[i], nullptr);
    }
    cleanup_();
}


void Renderer3D::updateUniformBuffer() {
    static auto startTime = std::chrono::high_resolution_clock::now();
    auto currentTime = std::chrono::high_resolution_clock::now();
    float time = std::chrono::duration<float, std::chrono::seconds::period>(currentTime - startTime).count();

    // TODO use push constant instead of ubo
    UniformBufferObject ubo{};
    /* ubo.model = glm::rotate(glm::mat4(1.0f), time * std::numbers::pi_v<float> / 2, glm::vec3(0.0f, 0.0f, 1.0f)); */
    /* ubo.view = glm::lookAt(glm::vec3(2.0f, 2.0f, 2.0f), glm::vec3(0.0f, 0.0f, 0.0f), glm::vec3(0.0f, 0.0f, 1.0f)); */
    /* ubo.projection = glm::perspective(glm::radians(45.0f), static_cast<float>(scExtent.width) / scExtent.height, 1.0f, 10.0f); */
    ubo.model = glm::mat4(1);
    ubo.view = glm::mat4(1);
    ubo.projection = glm::mat4(1);
    /* ubo.projection[1][1] *= -1;  // y coordinate inverted in opengl */
    void* data;
    vkMapMemory(vk.device, uniformBuffersMemory[vk.currentFrame], NO_OFFSET, sizeof(ubo), NO_FLAGS, &data);
    memcpy(data, &ubo, sizeof(ubo));
    vkUnmapMemory(vk.device, uniformBuffersMemory[vk.currentFrame]);
}


void Renderer3D::recordCommandBuffer(uint32_t imageIndex, uint32_t currentFrame) {
    VkCommandBufferBeginInfo commandBufferBI{};
    commandBufferBI.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
    /* commandBufferBI.flags = 0; */
    /* commandBufferBI.pInheritanceInfo = nullptr; */
    VkResult result = vkBeginCommandBuffer(commandBuffers[currentFrame], &commandBufferBI); 
    if (result != VK_SUCCESS) {
        throw getVkException(result, "Failed to begin 3D - command buffer", "recordCommandBuffer");
    }

    VkRenderPassBeginInfo renderPassBI{};
    renderPassBI.sType = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO;
    renderPassBI.renderPass = vk.renderPassBegin;
    // TODO
    /* renderPassBI.framebuffer = vk.scFramebuffers[0][imageIndex]; */
    renderPassBI.renderArea.offset = { 0, 0 };
    renderPassBI.renderArea.extent = vk.scExtent;
    std::array<VkClearValue, 2> clearValues{};
    clearValues[0].color = {{0.0f, 0.0f, 0.0f, 1.0f}};
    clearValues[1].depthStencil = {1.0f, 0};
    renderPassBI.clearValueCount = static_cast<uint32_t>(clearValues.size());
    renderPassBI.pClearValues = clearValues.data();

    vkCmdBeginRenderPass(commandBuffers[currentFrame], &renderPassBI, VK_SUBPASS_CONTENTS_INLINE);

    vkCmdBindPipeline(commandBuffers[currentFrame], VK_PIPELINE_BIND_POINT_GRAPHICS, vk.pipelines[PL_3D].pipeline);
    VkBuffer vertexBuffers[] = { vertexBuffer };
    VkDeviceSize offsets[] = {0};
    uint32_t bindingCount = 1;
    vkCmdBindVertexBuffers(commandBuffers[currentFrame], BINDING, bindingCount, vertexBuffers, offsets);
    // TODO use correct index type!
    vkCmdBindIndexBuffer(commandBuffers[currentFrame], indexBuffer, NO_OFFSET, VK_INDEX_TYPE_UINT32);

    uint32_t descriptorCount = 1;
    uint32_t firstSet = 0;
    uint32_t dynamicOffsetCount = 0;
    uint32_t* dynamicOffsets = nullptr;
    vkCmdBindDescriptorSets(commandBuffers[currentFrame], VK_PIPELINE_BIND_POINT_GRAPHICS, vk.pipelines[PL_3D].layout, firstSet, descriptorCount, &vk.descriptorSets[currentFrame],  dynamicOffsetCount, dynamicOffsets);

    int instanceCount = 1;
    int firstIndex = 0;
    int firstInstance = 0;
    /* vkCmdDrawIndexed(commandBuffers[currentFrame], static_cast<uint32_t>(shapesIndicesCount), instanceCount, firstIndex, NO_OFFSET, firstInstance); */

    vkCmdEndRenderPass(commandBuffers[currentFrame]);
    result = vkEndCommandBuffer(commandBuffers[currentFrame]); 
    if (result != VK_SUCCESS) {
        rLog.error("Failed to record command buffer", "VkResult:", STR_VK_RESULT(result));
        throw getVkException(result, "Failed to record command buffer");
    }
}



}  // namespace gz::vk
