#include "vertex.hpp"

#include <optional>
#include <unordered_map>
#include <vulkan/vulkan_core.h>

namespace gz::vk {
    enum PipelineT {
        PL_3D, PL_2D
    };
    struct Pipeline {
        VkPipeline pipeline;
        VkPipelineLayout layout;
        void operator=(const Pipeline& other) = delete;
    };

    /**
     * @brief Map holding pipelines
     */
    class PipelineContainer {
        public:
            using iterator = std::unordered_map<PipelineT, Pipeline>::iterator;
            Pipeline& operator[](const PipelineT& key) { return pipelines[key]; }
            /**
             * @brief Destroy the pipeline+layout and then remove the handles
             */
            void erase(const PipelineT& key, VkDevice& device, const VkAllocationCallbacks* pAllocator=nullptr);
            iterator erase(const iterator& it, VkDevice& device, const VkAllocationCallbacks* pAllocator=nullptr);
            iterator begin() { return pipelines.begin(); }
            iterator end() { return pipelines.end(); }
            size_t size() const { return pipelines.size(); }
        private:
            std::unordered_map<PipelineT, Pipeline> pipelines;

    };


    template<typename T>
    concept SupportedIndexType = std::same_as<T, uint16_t> or std::same_as<T, uint32_t>;

    template<SupportedIndexType T>
    struct VerticesAndIndices {
        std::vector<Vertex3D> vertices;
        std::vector<T> indices;
        constexpr VkIndexType getIndexType() const {
            if (std::same_as<T, uint16_t>) {
                return VK_INDEX_TYPE_UINT16;
            }
            else if (std::same_as<T, uint32_t>) {
                return VK_INDEX_TYPE_UINT32;
            }
        }
    };


    struct QueueFamilyIndices {
        std::optional<uint32_t> graphicsFamily;
        std::optional<uint32_t> presentFamily;
        std::optional<uint32_t> transferFamily;

        bool hasNecessaryValues() {
            return graphicsFamily.has_value() && presentFamily.has_value();
        }
        bool hasAllValues() {
            return graphicsFamily.has_value() && presentFamily.has_value() && transferFamily.has_value();
        }

        std::string toString() const {
            std::string s = "[ ";
            s += "graphicsFamily: ";
            if (graphicsFamily.has_value()) { s += std::to_string(graphicsFamily.value()); }
            else { s += "not set"; }
            s += ", presentFamily: ";
            if (presentFamily.has_value()) { s += std::to_string(presentFamily.value()); }
            else { s += "not set"; }
            s += ", transferFamily: ";
            if (transferFamily.has_value()) { s += std::to_string(transferFamily.value()); }
            else { s += "not set"; }
            return s + " ]";
        }
    };


    struct SwapChainSupport {
        VkSurfaceCapabilitiesKHR capabilities;
        std::vector<VkSurfaceFormatKHR> formats;
        std::vector<VkPresentModeKHR> presentModes;
    };
} // namespace gz::vk
